"""General interfaces for hardware related concepts.

It contains:
* the machines able to execute instructions and memories;
* the different types of memory.

"""
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Tuple
from z3 import BitVecRef, ExprRef


class IMemory(ABC):
    """Abstract class to represent a memory."""

    @abstractmethod
    def __init__(self, word_size: int = 8, symbolic: bool = False):
        """Initialize the memory.

        {wordsize} is the size of a word in bits.
        """
        self._word_size = word_size

    @abstractmethod
    def read(self, pos: int = 0) -> Tuple[BitVecRef, BitVecRef]:
        """Read the memory at pos {pos}.

        Return a tuple standing for the value and the true size of the data stored at
        position = {pos}.
        """
        pass

    @abstractmethod
    def store(self, val: Any, pos: int = None) -> None:
        """Store some value {val} in the memory at pos {pos}."""
        pass


class ISymbolicMemory(IMemory):
    """Abstract class for a symbolic memory.

    This follows the same interface than the concrete memory. Its purpose is to
    distinguish concrete from symbolic memories in the inheritance graph.

    Additionally, a symbolic memory needs to support a few more operations.
    """

    @abstractmethod
    def add_constraints(self, constraints: List[ExprRef]):
        """Add constraints to the memory."""
        pass

    @abstractmethod
    def merge(self, other, positive_timer: int, negative_timer: int):
        """Merge the content of the memory with another.

        The merging affects only the records whose `timer` is such
            * `negative_timer < timer`
            * or `timer > positive_timer`
        """
        pass


class IMachine(ABC):
    """Abstract class to represent a machine."""

    @abstractmethod
    def __init__(self, memory_constructor: Callable[[Dict[str, Any]], IMemory]):
        """Initialize the machine with type of memory.

        {memory_constructor} is a function that returns an object of type `IMemory`.
        """
        pass

    @property
    @abstractmethod
    def pc(self) -> int:
        """The program counter."""
        pass

    @pc.setter
    @abstractmethod
    def pc(self, new_pc: int) -> int:
        """Alters the program counter."""
        pass

    @abstractmethod
    def is_true(self, expr) -> bool:
        """Evaluate an expression and return a boolean."""
        pass


class IStackBasedMachine(IMachine):
    """Abstract class to represent a stack based machine.

    This machine can include several stacks. The default config includes one stack.
    """

    @abstractmethod
    def __init__(
        self,
        memory_constructor: Callable[[Dict[str, Any]], IMemory],
        stack_count: int = 1,
    ):
        """Initialize the stack machine.

        It requires a stack count and a type of memory to initialize.
        """
        pass

    @abstractmethod
    def push_onto_stack(self, element: Any, n: int = 0):
        """Push an element onto the stack {n}."""
        pass

    @abstractmethod
    def pop_from_stack(self, n: int = 0) -> Tuple[Any, Any]:
        """Pop and return an element from the stack {n}.

        The size of the element is passed along the value.

        Note: we need to pass the size in case we deal with symbols. In such a
        situation, the size might not be known and could be encoded with a symbol too.
        """
        pass

    @abstractmethod
    def drop_from_stack(self, drop_n: int, n: int = 0) -> int:
        """Drop {drop_n} elements from the stack {n} and return the new position."""
        pass

    @abstractmethod
    def count_stack_elements(self, n: int = 0) -> int:
        """Return the number of elements on the stack {n}."""
        pass


class IRAMBasedMachine(IMachine):
    """Abstract class to represent a RAM based machine."""

    @abstractmethod
    def read(self, address) -> Tuple[Any, Any]:
        """Read and return the value as position {address}.

        The size of the element is passed along the value.
        """
        pass


class ISymbolicMachine(IMachine):
    """Abstract class to represent a machine with symbolic capabilities.

    A symbolic machine must be able to merge with another machine.
    """

    @abstractmethod
    def merge(self, other, other_constraints) -> None:
        """Merge the another machine with the current one."""
        pass

    @abstractmethod
    def add_constraints(self, constraints: List[ExprRef]) -> None:
        """Add a new list of constraints to the current path."""
        pass
