"""Implementation of a stack-based machine able to run Bitcoin scripts."""
from copy import deepcopy
import hashlib
from math import ceil, floor, log2
from typing import Any, Callable, Dict, List, Tuple
from symbolic_execution.hardware.memory.ram import Word
from symbolic_execution.program.instruction.interfaces import IInstruction
from z3 import (
    BitVec,
    is_int,
    IntVal,
    is_int_value,
    Int,
    simplify,
    is_true,
    is_false,
    is_expr,
    is_bool,
    If,
    IntSort,
    BoolSort,
    BitVecSort,
    RecAddDefinition,
    And,
    RecFunction,
    BitVecRef,
    Or,
    ExprRef,
    Function,
    BoolVal,
)

from ..interfaces import (
    IMemory,
    IRAMBasedMachine,
    IStackBasedMachine,
    ISymbolicMachine,
    ISymbolicMemory,
)

# Btc script supports byte vectors up to 520 bytes
BTC_VECTOR_SIZE = 520 * 8


def str_to_int(hex: str) -> int:
    """Bitcoin has a custom way to convert an hex string to an int.

    Example:
        * 0x81 represents -1
        * 0x80 represents the negative 0

    See https://en.bitcoin.it/wiki/Script
    """
    hex = list(hex)

    if len(hex) <= 0:
        raise ValueError("Empty string.")
    if len(hex) % 2:
        raise ValueError("Invalid hex string: the numbers of characters is not even.")

    most_significant_byte = int(hex[-2], 16)
    sign_bit = most_significant_byte // 8

    # Unset the sign bit if set
    hex[-2] = f"{most_significant_byte & 0x7:0x}"

    abs_value = int.from_bytes(bytes.fromhex("".join(hex)), byteorder="little")
    return -abs_value if sign_bit else abs_value


def int_to_bytes(val: int) -> str:
    """Bitcoin has a custom way to represent an integer as a hex string.

    # NOTE: 0 is encoded as an empty string

    Example:
        * 0x81 represents -1
        * 0x80 represents the negative 0

    See https://en.bitcoin.it/wiki/Script
    """
    abs_val = abs(val)

    # BTC doesn't use complement of 2, over 1 byte the max encode value
    # belongs to [-127, 127]
    array_size = ceil((abs_val.bit_length() + 1) / 8)

    b = bytearray(abs_val.to_bytes(array_size, "little"))

    if val < 0:
        b[-1] = b[-1] | 0x80

    return b


# Set of functions specific to BTC
CHECK_SINGLE_SIG = Function("check_sig", IntSort(), IntSort(), BoolSort())
CHECK_MULTI_SIG = Function("check_multi_sig", IntSort(), IntSort(), BoolSort())
HASH_FUNCTIONS = {
    "ripemd160": [Function("ripemd160", IntSort(), IntSort()), hashlib.new, 20],
    "sha1": [Function("sha1", IntSort(), IntSort()), hashlib.new, 20],
    "sha256": [Function("sha256", IntSort(), IntSort()), hashlib.new, 32],
}


class BTCVector:
    """Bitcoin byte vector.

    Same as a word, with overloaded operators.
    """

    # Define a few constants
    BTC_TRUE = IntVal(1)
    BTC_FALSE = IntVal(0)

    @classmethod
    def _min_bytes_needed(cls, val: ExprRef):
        return If(
            val == 0,
            IntVal(0),
            If(
                And(val <= 2 ** 7 - 1, val >= -(2 ** 7 - 1)),
                IntVal(1),
                If(
                    And(val <= 2 ** 15 - 1, val >= -(2 ** 15 - 1)),
                    IntVal(2),
                    If(
                        And(val <= 2 ** 23 - 1, val >= -(2 ** 23 - 1)),
                        IntVal(3),
                        If(
                            And(val <= 2 ** 31 - 1, val >= -(2 ** 31 - 1)),
                            IntVal(4),
                            IntVal(5),
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, val: BitVecRef, size: BitVecRef):
        """Initialize the BTC vector.

        * {val} is a bit vector of {BTC_VECTOR_SIZE} bits.
        * {size} is a bit vector of {BTC_VECTOR_SIZE} bits.

        """
        if not is_int(val):
            raise ValueError("Invalid value to initialize a BTC vector.")
        if not is_int(size):
            raise ValueError("Invalid value to initialize a BTC vector.")

        self.val = val
        self.size = size

    def __add__(self, other):
        """Return the sum of 2 btc vectors."""
        return BTCVector(
            self.val + other.val, self._min_bytes_needed(self.val + other.val)
        )

    def __sub__(self, other):
        """Return the difference of 2 btc vectors."""
        return BTCVector(
            self.val - other.val, self._min_bytes_needed(self.val - other.val)
        )

    def __neg__(self):
        """Return the negation of the btc vector."""
        return BTCVector(-self.val, self.size)

    def __abs__(self):
        """Return the absolute value of the btc vector."""
        return BTCVector(If(self.val < 0, -self.val, self.val), self.size)

    def __eq__(self, other):
        """Return the BTC vector `True` if `self.val == other.val`."""
        if isinstance(other, int):
            other = BTCVector(IntVal(other), self._min_bytes_needed(other))

        # We can't find values such the hash is equal to some values
        if self.is_result_of_hash() or other.is_result_of_hash():
            return BTCVector(self.BTC_FALSE, IntVal(0))

        return BTCVector(
            If(
                And(self.val == other.val, self.size == other.size),
                self.BTC_TRUE,
                self.BTC_FALSE,
            ),
            If(
                And(self.val == other.val, self.size == other.size),
                IntVal(1),
                IntVal(0),
            ),
        )

    def __le__(self, other):
        """Return the BTC vector `True` if `self.val <= other.val`."""
        return BTCVector(
            If(self.val <= other.val, self.BTC_TRUE, self.BTC_FALSE),
            If(self.val <= other.val, IntVal(1), IntVal(0)),
        )

    def __lt__(self, other):
        """Return the BTC vector `True` if `self.val < other.val`."""
        return BTCVector(
            If(self.val < other.val, self.BTC_TRUE, self.BTC_FALSE),
            If(self.val < other.val, IntVal(1), IntVal(0)),
        )

    def __gt__(self, other):
        """Return the BTC vector `True` if `self.val > other.val`."""
        return BTCVector(
            If(self.val > other.val, self.BTC_TRUE, self.BTC_FALSE),
            If(self.val > other.val, IntVal(1), IntVal(0)),
        )

    def __ge__(self, other):
        """Return the BTC vector `True` if `self.val >= other.val`."""
        return BTCVector(
            If(self.val >= other.val, self.BTC_TRUE, self.BTC_FALSE),
            If(self.val >= other.val, IntVal(1), IntVal(0)),
        )

    def __and__(self, other):
        """Return the BTC vector `True` if `self.val != 0 and other.val != 0`."""
        return BTCVector(
            If(And(self.val != 0, other.val != 0), self.BTC_TRUE, self.BTC_FALSE),
            If(And(self.val != 0, other.val != 0), IntVal(1), IntVal(0)),
        )

    def __or__(self, other):
        """Return the BTC vector `True` if `self.val != 0 or other.val != 0`."""
        return BTCVector(
            If(Or(self.val != 0, other.val != 0), self.BTC_TRUE, self.BTC_FALSE),
            If(Or(self.val != 0, other.val != 0), IntVal(1), IntVal(0)),
        )

    def __min__(self, other):
        """Return the smallest BTC vector between {self} and {other}."""
        return BTCVector(
            If(self.val < other.val, self.val, other.val),
            If(self.val < other.val, self.size, other.size),
        )

    def __max__(self, other):
        """Return the smallest BTC vector between {self} and {other}."""
        return BTCVector(
            If(self.val > other.val, self.val, other.val),
            If(self.val > other.val, self.size, other.size),
        )

    def __within__(self, a, b):
        """Return the BTC vector `True` if `a.val <= self.val < b.val`."""
        return BTCVector(
            If(And(self.val < b.val, self.val >= a.val), self.BTC_TRUE, self.BTC_FALSE),
            If(And(self.val < b.val, self.val >= a.val), IntVal(1), IntVal(0)),
        )

    def __inc__(self):
        """Return self + 1."""
        return self + BTC_VECTOR_1

    def __dec__(self):
        """Return self - 1."""
        return self - BTC_VECTOR_1

    def __not__(self):
        """If the input is 0 or 1, it is flipped. Otherwise the output will be 0."""
        return BTCVector(
            If(self.val == 0, self.BTC_TRUE, self.BTC_FALSE),
            If(self.val == 0, IntVal(1), IntVal(0)),
        )

    def __num_eq__(self, other):
        """Return the BTC vector `True` if `self.val == other.val`."""
        return BTCVector(
            If(self.val == other.val, self.BTC_TRUE, self.BTC_FALSE),
            If(self.val == other.val, IntVal(1), IntVal(0)),
        )

    def __size__(self):
        """Return the number of bytes the vector is composed of."""
        return BTCVector(self.size, self._min_bytes_needed(self.size))

    def __hash__(hash_function_name: str):
        """Hash decorator that returns a function that computes the hash of the vector."""

        z3_function = HASH_FUNCTIONS[hash_function_name][0]
        py_function = HASH_FUNCTIONS[hash_function_name][1]
        hash_size = HASH_FUNCTIONS[hash_function_name][2]

        def hash_of_vector(btcVector):
            if is_int_value(btcVector.val):
                m = py_function(hash_function_name, int_to_bytes(simplify(btcVector.val).as_long()))
                h = m.hexdigest()
                return BTCVector(IntVal(str_to_int(h)), IntVal(hash_size))
            else:
                f = z3_function
                return BTCVector(f(btcVector.val), IntVal(hash_size))

        return hash_of_vector

    def __identity__(self):
        """Return self"""
        return self

    def check_sig(pubkeys: list, signatures: list):
        """Check a signature against a public key."""
        # TODO Could be refined. We could compute the signature in case both signature
        # and key are constants.
        if len(pubkeys) != 1 and len(signatures) != 1:
            raise ValueError("One public key and one signature are expected.")
        ret = CHECK_SINGLE_SIG(pubkeys[0].val, signatures[0].val)

        return BTCVector(
            If(ret == BoolVal(False), BTCVector.BTC_FALSE, BTCVector.BTC_TRUE),
            If(ret == BoolVal(False), IntVal(0), IntVal(1)),
        )

    def check_multisig(pubkeys_cnt, sigs_cnt, pubkeys, sigs):
        """Check a list of signatures against a list of public keys."""
        if pubkeys_cnt.val == 0:
            return BTCVector(BTCVector.BTC_TRUE, IntVal(1))

        pubkeys_unset = 0
        for pubkey in pubkeys:
            const_values = pubkey.fetch_const_values()
            pubkeys_unset = (
                pubkeys_unset + 1 if len(const_values) == 0 else pubkeys_unset
            )

        unsafe = False

        # If the number of public keys is not hardcoded, the attacker could forge it
        # and pass controlled keys
        if len(pubkeys_cnt.fetch_const_values()) == 0:
            unsafe = True
        #  If the number of required signatures is not set or lower than the number of
        # hardcoded public keys, the attacked could sign it correctly.
        if len(sigs_cnt.fetch_const_values()) == 0 or (
            is_int_value(sigs_cnt.val) and sigs_cnt.val.as_long() <= pubkeys_unset
        ):
            unsafe = True

        if unsafe:
            ret = CHECK_MULTI_SIG(pubkeys_cnt.val, sigs_cnt.val)
            return BTCVector(
                If(ret == BoolVal(False), BTCVector.BTC_FALSE, BTCVector.BTC_TRUE),
                If(ret == BoolVal(False), IntVal(0), IntVal(1)),
            )

        return BTCVector(BTCVector.BTC_FALSE, IntVal(0))

    def is_result_of_hash(self):
        def parse_ast(e):
            if e.decl().name() == "if":
                return parse_ast(e.arg(1)) or parse_ast(e.arg(2))
            else:
                z3_hash_functions = list(map(lambda v: v[0], HASH_FUNCTIONS.values()))
                return e.decl() in z3_hash_functions

        return parse_ast(self.val)

    def fetch_const_values(self):
        def parse_ast(e):
            values = []

            if is_int_value(e):
                return [e]
            elif e.decl().name() == "if":
                values.extend(parse_ast(e.arg(1)))
                values.extend(parse_ast(e.arg(2)))
                return values

            return values

        return parse_ast(self.val)


BTC_VECTOR_1 = BTCVector(IntVal(1), BTCVector._min_bytes_needed(1))


class SymbolicConditionError(Exception):
    """A symbolic condition couldn't be evaluated.

    Raised by the machine when a symbolic expression can't be evaluated
    to True or False.
    """

    def __init__(self, message, condition):
        """Initialize the error with the condition which triggered it."""
        self.condition = condition
        super().__init__(message)


class BTCMachine(IStackBasedMachine, IRAMBasedMachine, ISymbolicMachine):
    """Class to represent a Bitcoin machine able to run a script."""

    @property
    def stack_count(self) -> int:
        """Return the number of stacks implemented on a machine."""
        return self._stack_count

    def __init__(
        self,
        memory_constructor: Callable[[Dict[str, Any]], IMemory],
        word_size: int,
        stack_count: int = 2,
    ):
        """Initialize the BTC machine.

        It supports 2 stacks.
        """
        self.__memory_init(memory_constructor, word_size, stack_count)

        # Program counter
        self._pc = 0

        self._constraints = True

    def __memory_init(self, memory_constructor, word_size, stack_count):
        if stack_count != 2:
            raise ValueError("A BTC machine must include 2 stacks.")
        self._stack_count = stack_count
        self._stacks = [
            memory_constructor(word_size=word_size, mem_id=f"stack{i}")
            for i in range(stack_count)
        ]
        self._word_size = word_size
        # Because this memory is allowed to generate new symbols, the stack can't be
        # considered empty even no element have been pushed.
        if isinstance(self._stacks[0], ISymbolicMemory):
            self._stack_sizes = [
                Int(f"initial_element_count_on_stack_{i}") for i in range(stack_count)
            ]
        else:
            self._stack_sizes = [IntVal(0) for _ in range(stack_count)]

        # Memory position: btc memory works as a stack, keep track of the top of the mem
        # In case of symbolic execution, all paths are explored in parallel and merged
        # later. Thus, the position needs to be symbolic too.
        self._mem_pos = [IntVal(0) for _ in range(stack_count)]

    def __convert_val_to_word(self, val) -> Word:
        """Convert a value to a word.

        A word is a pair of symbols that represent the value and the size occupied by
        the value on the word.
        """
        if isinstance(val, bool):
            val = simplify(If(val, IntVal(1), IntVal(0)))
            size = IntVal(1)
        elif isinstance(val, int):
            if val < -(2 ** 31 - 1) or val > (2 ** 31 - 1):
                raise ValueError(f"This integer is too large to be encoded: {val}.")
            val = IntVal(val)
            size = BTCVector._min_bytes_needed(val)
        elif isinstance(val, str):
            bytes_needed = len(val) // 2  # The string must be in hex format
            if bytes_needed > self._word_size // 8:
                raise ValueError(
                    f"A BTC vector is composed of at max {self._word_size // 8} bytes."
                )

            val = str_to_int(val)
            size = IntVal(bytes_needed)
        elif isinstance(val, BTCVector):
            size = val.size
            val = val.val
        else:
            raise NotImplementedError(
                f"Support for this type {type(val)} has not been implemented yet."
            )

        return Word(val, size)

    # ################ #
    # IMachine Methods #
    # ################ #
    @property
    def pc(self) -> int:
        """The program counter."""
        return self._pc

    @pc.setter
    def pc(self, new_pc: int) -> int:
        """Alters the program counter."""
        self._pc = new_pc

    def is_true(self, top_elt: BTCVector) -> bool:
        """Evaluate an expression and return a boolean."""
        # In btc, False is represented by 0
        expr = simplify(top_elt.val != 0)

        true, false = is_true(expr), is_false(expr)

        if not true and not false:
            raise SymbolicConditionError(
                f"This expression can't be evaluated to True nor False: {expr}",
                top_elt.val != 0,
            )
        elif true and false:
            # This case should never happen
            raise ValueError(
                f"This expression is evaluated to True and False: {expr}",
                top_elt.val != 0,
            )

        return true

    # ########################## #
    # IStackBasedMachine Methods #
    # ########################## #

    def _check_nth_stack(function_to_decorate):
        """Check the stack exists."""

        def wrapper(machine, *args, **kwargs):
            if "n" in kwargs and kwargs["n"] >= machine._stack_count:
                raise IndexError(
                    f"This machine is composed of {machine._stack_count} stacks, thus "
                    f"the stack #{kwargs['n']} doesn't exist."
                )

            return function_to_decorate(machine, *args, **kwargs)

        return wrapper

    @_check_nth_stack
    def push_onto_stack(self, element: Any, n: int = 0):
        """Put an element onto the stack {n}."""
        self._stacks[n].store(self.__convert_val_to_word(element), pos=self._mem_pos[n])
        self._mem_pos[n] += 1
        self._stack_sizes[n] += 1

    @_check_nth_stack
    def pop_from_stack(self, n: int = 0) -> BTCVector:
        """Pop and return an element from the stack {n}.

        The size of the element is passed along the value.
        """
        self._mem_pos[n] -= 1
        top_word = self._stacks[n].read(pos=self._mem_pos[n])
        self._stack_sizes[n] -= 1
        return BTCVector(top_word.val, top_word.size)

    @_check_nth_stack
    def drop_from_stack(self, drop_n: int, n: int = 0) -> int:
        """Drop {drop_n} elements from the stack {n}."""
        self._mem_pos[n] -= drop_n.val

        return self._mem_pos[n]

    @_check_nth_stack
    def count_stack_elements(self, n: int = 0) -> BTCVector:
        """Return the number of elements on the stack {n}."""
        return BTCVector(
            self._stack_sizes[n], BTCVector._min_bytes_needed(self._stack_sizes[n])
        )

    # ######################## #
    # IRAMBasedMachine Methods #
    # ######################## #

    def read(self, address: BTCVector) -> Tuple[Any, Any]:
        """Read and return the value x hops from the top of the stack.

        The size of the element is passed along the value.
        """
        # Arbitrary read are only allowed on the main stack.
        elt = self._stacks[0].read(pos=self._mem_pos[0] - address.val)

        return BTCVector(elt.val, elt.size)

    # ######################## #
    # ISymbolicMachine Methods #
    # ######################## #

    def merge(self, other, other_constraints: List[ExprRef]) -> None:
        """Create a new machine resulting from the merge of 2 machines."""
        other_constraints = And(other._constraints, *other_constraints)

        if other is None:
            return self

        if self._stack_count != other._stack_count:
            raise ValueError(
                "The 2 machines to merge do not have the same number of stacks."
            )
        if self._word_size != other._word_size:
            raise ValueError("The 2 machines to merge do not have the same word size.")

        for i in range(self._stack_count):
            self._stack_sizes[i] = simplify(
                If(And(self._constraints), self._stack_sizes[i], other._stack_sizes[i])
            )
            self._mem_pos[i] = simplify(
                If(And(self._constraints), self._mem_pos[i], other._mem_pos[i])
            )
            self._stacks[i].merge(
                other._stacks[i], 0, 0, self._constraints, other_constraints
            )

        self._constraints = Or(self._constraints, other_constraints)

        return self

    def add_constraints(self, constraints: List[ExprRef]) -> None:
        """Add a new constraint to the current path."""
        self._constraints = And(self._constraints, *constraints)
        for mem in self._stacks:
            mem.add_constraints(constraints)
