"""Code by Vincent Jacquot - University of Liège."""
import operator
from typing import Any, Callable, Dict, List, Tuple
from symbolic_execution.language.exceptions import InvalidOpCode
from symbolic_execution.program.instruction.general import (
    NOP,
    InvalidInstruction,
    InstructionTag,
    UnconditionalJump,
)
from symbolic_execution.program.instruction.hybrid import (
    CheckMultiSigInstruction,
    PickInstruction,
)

from symbolic_execution.program.instruction.interfaces import IInstruction
from symbolic_execution.program.instruction.stack_based import (
    ArithmeticOperator,
    CheckSigInstruction,
    ConditionalJump,
    PushDataOperator,
    StackAlterationOperator,
    SwapDataOperator,
)
from symbolic_execution.hardware.machines.btc_machine import (
    CHECK_SINGLE_SIG,
    BTCVector,
    BTCMachine,
)
from z3 import And, Int, Or, Function, IntSort, BoolVal

# Special Bitcoin Fields
nLockTime = Int("nLockTime")
nSequence = Int("nSequence")
version = Int("version")


def arithmetic_constraint_on_operands_size(ops: List[BTCVector]):
    """Return a constraint on an arithmetic operand to enforce its minimal encoding.

    An operand is minimally encoded if it is represented with the least amount of bytes.

    Example:
        01 is the minimal hex representation of 1.
        0100 is not a minimal hex representation of 1.
    """
    constraints = [op.size == BTCVector._min_bytes_needed(op.val) for op in ops]
    return constraints


def locktime_constraint(ops: List[BTCVector]):
    """Return a constraint to mimic OP_CHECKLOCKTIMEVERIFY specification.

    The conditions are:
        * top stack item <= the transaction's nLockTime field;
        * top stack item >= 0
        * If the transaction's nLockTime field >= 500000000,
          then top stack item >= 500000000
        * the input's nSequence field != 0xffffffff.
    while it must stay positive.
    """
    LOCKTIME_THRESHOLD = 500000000

    if len(ops) != 1:
        raise ValueError("Sequence operator should take one arg.")

    op = ops[0]

    condition = And(
        op.val <= nLockTime,
        op.val >= 0,
        Or(
            And(nLockTime > LOCKTIME_THRESHOLD, op.val >= LOCKTIME_THRESHOLD),
            And(nLockTime < LOCKTIME_THRESHOLD, op.val < LOCKTIME_THRESHOLD),
        ),
        nSequence != 0xFFFFFFFF,
    )

    return [condition]


def sequence_constraint(ops: List[BTCVector]):
    """Return a constraint to mimic OP_CHECKSEQUENCEVERIFY specification.

    The conditions are:
        * the top item is >= 0
        * the top item has the disable flag (1 << 31) set; or
            * the transaction version >= 2; and
            * the transaction input sequence number disable flag (1 << 31) is unset; and
            * the relative lock-time type is the same; and
            * the top stack item is smaller or equal than the
              transaction input sequence (when masked according to the BIP68);

    Note: we don't need to implement the condition relative to the relative locktime,
    because an unsafe script that can't be unlocked right now should still be considered
    unsafe.

    See https://github.com/bitcoin/bips/blob/master/bip-0112.mediawiki
    """

    def disable_flag(val):
        """Return the 32nd bit of the vector."""
        return (val % (2 ** 32)) / (2 ** 31)

    def lock_time_type_flag(val):
        """Return the 23th bit of the vector."""
        return (val % (2 ** 23)) / (2 ** 22)

    if len(ops) != 1:
        raise ValueError("Sequence operator should take one arg.")

    op = ops[0]

    condition = And(
        op.val >= 0,
        Or(
            And(
                version >= 2,
                disable_flag(nSequence) == 0,
                lock_time_type_flag(op.val) == lock_time_type_flag(nSequence),
                op.val <= nSequence,
            )
        ),
    )

    return [condition]


def signature_constraints(ops: List[BTCVector]):
    """Enforces the logic that no valid signature can be found for a given key."""
    if len(ops) != 2:
        raise ValueError("Exactly one pubkey and one signature are expected.")

    pubkeys, sigs = ops
    conditions = []

    for pubkey in pubkeys:
        for sig in sigs:
            for pk_const_val in pubkey.fetch_const_values():
                conditions.append(
                    CHECK_SINGLE_SIG(pk_const_val, sig.val) == BoolVal(False)
                )

    return conditions


class OpCode:
    """Represent an opcode in a language."""

    def __init__(
        self,
        token: str,
        instruction_constructor: Callable[[Dict[str, Any]], IInstruction] = None,
        branching: bool = False,
        default_args: dict = None,
    ):
        """Initialize an opcode.

        In particular,
            * {byte} is its integer representation;
            * {name} is its string representation;
            * {instruction_constructor} is a constructor that returns an
                instruction.
            * {branching} is a boolean indicating if the opcode is a branching
                operator.
            * {default_args} the args to be passed along with any call to
                {instruction_constructor}
        """
        self.token = token
        self.instruction_constructor = instruction_constructor
        self.branching = branching
        self.default_args = default_args if default_args else {}


class OpCodes:
    """Different types of scripts symbols for BTC."""

    __int_to_tokens = {
        0: "OP_FALSE",
        0x4F: "OP_1NEGATE",
        0x50: "OP_RESERVED",
        **{0x50 + i: f"OP_{i}" for i in range(1, 17)},
        0x61: "OP_NOP",
        0x62: "OP_VER",
        0x63: "OP_IF",
        0x64: "OP_NOT OP_IF",
        0x65: "OP_VERIF",
        0x66: "OP_VERNOTIF",
        0x67: "OP_ELSE",
        0x68: "OP_ENDIF",
        0x69: "OP_VERIFY",
        0x6A: "OP_RETURN",
        0x6B: "OP_TOALTSTACK",
        0x6C: "OP_FROMALTSTACK",
        0x6D: "OP_2DROP",
        0x6E: "OP_2DUP",
        0x6F: "OP_3DUP",
        0x70: "OP_2OVER",
        0x71: "OP_2ROT",
        0x72: "OP_2SWAP",
        0x73: "OP_DUP OP_IF OP_DUP OP_ENDIF",
        0x74: "OP_DEPTH",
        0x75: "OP_DROP",
        0x76: "OP_DUP",
        0x77: "OP_NIP",
        0x78: "OP_OVER",
        0x79: "OP_PICK",
        0x7A: "OP_ROLL",
        0x7B: "OP_ROT",
        0x7C: "OP_SWAP",
        0x7D: "OP_TUCK",
        0x7E: "OP_CAT",
        0x7F: "OP_SUBSTR",
        0x80: "OP_LEFT",
        0x81: "OP_RIGHT",
        # OP_SIZE is implemented as a 1-1 function: elt -> elt.size. However, Bitcoin
        # defines OP_SIZE as a 1-2 function: elt -> elt, elt.size
        0x82: "OP_DUP OP_SIZE",
        0x83: "OP_INVERT",
        0x84: "OP_AND",
        0x85: "OP_OR",
        0x86: "OP_XOR",
        0x87: "OP_EQUAL",
        0x88: "OP_EQUAL OP_VERIFY",
        0x89: "OP_RESERVED1",
        0x8A: "OP_RESERVED2",
        0x8B: "OP_1ADD",
        0x8C: "OP_1SUB",
        0x8D: "OP_2MUL",
        0x8E: "OP_2DIV",
        0x8F: "OP_NEGATE",
        0x90: "OP_ABS",
        0x91: "OP_NOT",
        0x92: "OP_FALSE OP_EQUAL OP_NOT",
        0x93: "OP_ADD",
        0x94: "OP_SUB",
        0x95: "OP_MUL",
        0x96: "OP_DIV",
        0x97: "OP_MOD",
        0x98: "OP_LSHIFT",
        0x99: "OP_RSHIFT",
        0x9A: "OP_BOOLAND",
        0x9B: "OP_BOOLOR",
        0x9C: "OP_NUMEQUAL",
        0x9D: "OP_NUMEQUAL OP_VERIFY",
        0x9E: "OP_NUMEQUAL OP_NOT",
        0x9F: "OP_LESSTHAN",
        0xA0: "OP_GREATERTHAN",
        0xA1: "OP_LESSTHANOREQUAL",
        0xA2: "OP_GREATERTHANOREQUAL",
        0xA3: "OP_MIN",
        0xA4: "OP_MAX",
        0xA5: "OP_WITHIN",
        0xA6: "OP_RIPEMD160",
        0xA7: "OP_SHA1",
        0xA8: "OP_SHA256",
        0xA9: "OP_SHA256 OP_RIPEMD160",
        0xAA: "OP_SHA256 OP_SHA256",
        0xAB: "OP_CODESEPARATOR",
        0xAC: "OP_CHECKSIG",
        0xAD: "OP_CHECKSIG OP_VERIFY",
        0xAE: "OP_CHECKMULTISIG",
        0xAF: "OP_CHECKMULTISIG OP_VERIFY",
        0xB0: "OP_NOP1",
        0xB1: "OP_DUP OP_CHECKLOCKTIME OP_VERIFY",
        0xB2: "OP_DUP OP_CHECKSEQUENCE OP_VERIFY",
        0xB3: "OP_NOP4",
        0xB4: "OP_NOP5",
        0xB5: "OP_NOP6",
        0xB6: "OP_NOP7",
        0xB7: "OP_NOP8",
        0xB8: "OP_NOP9",
        0xB9: "OP_NOP10",
        0xBA: "OP_CHECKSIGADD",
        0xFF: "OP_INVALIDOPCODE",
    }

    __token_to_op_codes = {
        "OP_FALSE": OpCode(
            "OP_FALSE",
            instruction_constructor=PushDataOperator,
            default_args={"data": 0},
        ),
        "OP_1NEGATE": OpCode(
            "OP_1NEGATE",
            instruction_constructor=PushDataOperator,
            default_args={"data": -1},
        ),
        "OP_RESERVED": OpCode(
            "OP_RESERVED",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_RESERVED", "tag": InstructionTag.RESERVED},
        ),
        **{
            f"OP_{i}": OpCode(
                f"OP_{i}",
                instruction_constructor=PushDataOperator,
                default_args={"data": i},
            )
            for i in range(1, 17)
        },
        # control
        "OP_NOP": OpCode("OP_NOP", instruction_constructor=NOP),
        "OP_VER": OpCode(
            "OP_VER",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_VER", "tag": InstructionTag.RESERVED},
        ),
        "OP_IF": OpCode(
            "OP_IF", branching=True, instruction_constructor=ConditionalJump
        ),
        "OP_VERIF": OpCode(
            "OP_VERIF",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_VERIF", "tag": InstructionTag.RESERVED},
        ),
        "OP_VERNOTIF": OpCode(
            "OP_VERNOTIF",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_VERNOTIF", "tag": InstructionTag.RESERVED},
        ),
        # Removed during second pass
        "OP_ELSE": OpCode(
            "OP_ELSE",
            branching=True,
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_ELSE"},
        ),
        # Removed during second pass
        "OP_ENDIF": OpCode(
            "OP_ENDIF",
            branching=True,
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_ENDIF"},
        ),
        "OP_VERIFY": OpCode(
            "OP_VERIFY", branching=True, instruction_constructor=ConditionalJump
        ),
        "OP_RETURN": OpCode(
            "OP_RETURN", branching=True, instruction_constructor=UnconditionalJump
        ),
        # stack ops
        "OP_TOALTSTACK": OpCode(
            "OP_TOALTSTACK",
            instruction_constructor=SwapDataOperator,
            default_args={"stack_from": 0, "stack_to": 1},
        ),
        "OP_FROMALTSTACK": OpCode(
            "OP_FROMALTSTACK",
            instruction_constructor=SwapDataOperator,
            default_args={"stack_from": 1, "stack_to": 0},
        ),
        "OP_2DROP": OpCode(
            "OP_2DROP",
            instruction_constructor=StackAlterationOperator,
            default_args={"pattern": (["x1", "x2"], [])},
        ),
        "OP_2DUP": OpCode(
            "OP_2DUP",
            instruction_constructor=StackAlterationOperator,
            default_args={"pattern": (["x1", "x2"], ["x1", "x2", "x1", "x2"])},
        ),
        "OP_3DUP": OpCode(
            "OP_3DUP",
            instruction_constructor=StackAlterationOperator,
            default_args={
                "pattern": (["x1", "x2", "x3"], ["x1", "x2", "x3", "x1", "x2", "x3"])
            },
        ),
        "OP_2OVER": OpCode(
            "OP_2OVER",
            instruction_constructor=StackAlterationOperator,
            default_args={
                "pattern": (
                    ["x1", "x2", "x3", "x4"],
                    ["x1", "x2", "x3", "x4", "x1", "x2"],
                )
            },
        ),
        "OP_2ROT": OpCode(
            "OP_2ROT",
            instruction_constructor=StackAlterationOperator,
            default_args={
                "pattern": (
                    ["x1", "x2", "x3", "x4", "x5", "x6"],
                    ["x3", "x4", "x5", "x6", "x1", "x2"],
                )
            },
        ),
        "OP_2SWAP": OpCode(
            "OP_2SWAP",
            instruction_constructor=StackAlterationOperator,
            default_args={
                "pattern": (["x1", "x2", "x3", "x4"], ["x3", "x4", "x1", "x2"])
            },
        ),
        "OP_DEPTH": OpCode(
            "OP_DEPTH",
            instruction_constructor=PushDataOperator,
            default_args={"data": BTCMachine.count_stack_elements},
        ),
        "OP_DROP": OpCode(
            "OP_DROP",
            instruction_constructor=StackAlterationOperator,
            default_args={"pattern": (["x1"], [])},
        ),
        "OP_DUP": OpCode(
            "OP_DUP",
            instruction_constructor=StackAlterationOperator,
            default_args={"pattern": (["x1"], ["x1", "x1"])},
        ),
        "OP_NIP": OpCode(
            "OP_NIP",
            instruction_constructor=StackAlterationOperator,
            default_args={"pattern": (["x1", "x2"], ["x2"])},
        ),
        "OP_OVER": OpCode(
            "OP_OVER",
            instruction_constructor=StackAlterationOperator,
            default_args={"pattern": (["x1", "x2"], ["x1", "x2", "x1"])},
        ),
        "OP_PICK": OpCode(
            "OP_PICK", instruction_constructor=PickInstruction, default_args={}
        ),
        # TODO implement
        "OP_ROLL": OpCode(
            "OP_ROLL",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_ROLL", "tag": InstructionTag.NOT_IMPLEMENTED},
        ),
        "OP_ROT": OpCode(
            "OP_ROT",
            instruction_constructor=StackAlterationOperator,
            default_args={"pattern": (["x1", "x2", "x3"], ["x2", "x3", "x1"])},
        ),
        "OP_SWAP": OpCode(
            "OP_SWAP",
            instruction_constructor=StackAlterationOperator,
            default_args={"pattern": (["x1", "x2"], ["x2", "x1"])},
        ),
        "OP_TUCK": OpCode(
            "OP_TUCK",
            instruction_constructor=StackAlterationOperator,
            default_args={"pattern": (["x1", "x2"], ["x2", "x1", "x2"])},
        ),
        # splice ops
        "OP_CAT": OpCode(
            "OP_CAT",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_CAT", "tag": InstructionTag.DISABLED},
        ),
        "OP_SUBSTR": OpCode(
            "OP_SUBSTR",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_SUBSTR", "tag": InstructionTag.DISABLED},
        ),
        "OP_LEFT": OpCode(
            "OP_LEFT",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_LEFT", "tag": InstructionTag.DISABLED},
        ),
        "OP_RIGHT": OpCode(
            "OP_RIGHT",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_RIGHT", "tag": InstructionTag.DISABLED},
        ),
        "OP_SIZE": OpCode(
            "OP_SIZE",
            instruction_constructor=ArithmeticOperator,
            default_args={"arity": 1, "operator": BTCVector.__size__, "symbol": "size"},
        ),
        # bit logic
        "OP_INVERT": OpCode(
            "OP_INVERT",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_RIGHT", "tag": InstructionTag.DISABLED},
        ),
        "OP_AND": OpCode(
            "OP_AND",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_RIGHT", "tag": InstructionTag.DISABLED},
        ),
        "OP_OR": OpCode(
            "OP_OR",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_RIGHT", "tag": InstructionTag.DISABLED},
        ),
        "OP_XOR": OpCode(
            "OP_XOR",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_RIGHT", "tag": InstructionTag.DISABLED},
        ),
        "OP_EQUAL": OpCode(
            "OP_EQUAL",
            instruction_constructor=ArithmeticOperator,
            default_args={"arity": 2, "operator": operator.eq, "symbol": "=="},
        ),
        "OP_RESERVED1": OpCode(
            "OP_RESERVED1",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_RESERVED1", "tag": InstructionTag.RESERVED},
        ),
        "OP_RESERVED2": OpCode(
            "OP_RESERVED2",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_RESERVED2", "tag": InstructionTag.RESERVED},
        ),
        # numeric
        "OP_1ADD": OpCode(
            "OP_1ADD",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": BTCVector.__inc__,
                "symbol": "++",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_1SUB": OpCode(
            "OP_1SUB",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": BTCVector.__dec__,
                "symbol": "--",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_2MUL": OpCode(
            "OP_2MUL",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_2MUL", "tag": InstructionTag.DISABLED},
        ),
        "OP_2DIV": OpCode(
            "OP_2DIV",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_2DIV", "tag": InstructionTag.DISABLED},
        ),
        "OP_NEGATE": OpCode(
            "OP_NEGATE",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": operator.neg,
                "symbol": "neg",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_ABS": OpCode(
            "OP_ABS",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": operator.abs,
                "symbol": "abs",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_NOT": OpCode(
            "OP_NOT",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": BTCVector.__not__,
                "symbol": "not",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_ADD": OpCode(
            "OP_ADD",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": operator.add,
                "symbol": "+",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_SUB": OpCode(
            "OP_SUB",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": operator.sub,
                "symbol": "-",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_MUL": OpCode(
            "OP_MUL",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_MUL", "tag": InstructionTag.DISABLED},
        ),
        "OP_DIV": OpCode(
            "OP_DIV",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_DIV", "tag": InstructionTag.DISABLED},
        ),
        "OP_MOD": OpCode(
            "OP_MOD",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_MOD", "tag": InstructionTag.DISABLED},
        ),
        "OP_LSHIFT": OpCode(
            "OP_LSHIFT",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_LSHIFT", "tag": InstructionTag.DISABLED},
        ),
        "OP_RSHIFT": OpCode(
            "OP_RSHIFT",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_RSHIFT", "tag": InstructionTag.DISABLED},
        ),
        "OP_BOOLAND": OpCode(
            "OP_BOOLAND",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": operator.and_,
                "symbol": "&",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_BOOLOR": OpCode(
            "OP_BOOLOR",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": operator.or_,
                "symbol": "|",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_NUMEQUAL": OpCode(
            "OP_NUMEQUAL",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": BTCVector.__num_eq__,
                "symbol": "num_eq",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_LESSTHAN": OpCode(
            "OP_LESSTHAN",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": operator.lt,
                "symbol": "<",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_GREATERTHAN": OpCode(
            "OP_GREATERTHAN",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": operator.gt,
                "symbol": ">",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_LESSTHANOREQUAL": OpCode(
            "OP_LESSTHANOREQUAL",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": operator.le,
                "symbol": "<=",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_GREATERTHANOREQUAL": OpCode(
            "OP_GREATERTHANOREQUAL",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": operator.ge,
                "symbol": ">=",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_MIN": OpCode(
            "OP_MIN",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": BTCVector.__min__,
                "symbol": "min",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_MAX": OpCode(
            "OP_MAX",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 2,
                "operator": BTCVector.__max__,
                "symbol": "max",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_WITHIN": OpCode(
            "OP_WITHIN",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 3,
                "operator": BTCVector.__within__,
                "symbol": "within",
                "constraint_on_operands": arithmetic_constraint_on_operands_size,
            },
        ),
        "OP_RIPEMD160": OpCode(
            "OP_RIPEMD160",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": BTCVector.__hash__("ripemd160"),
                "symbol": "ripemd160",
            },
        ),
        "OP_SHA1": OpCode(
            "OP_SHA1",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": BTCVector.__hash__("sha1"),
                "symbol": "sha1",
            },
        ),
        "OP_SHA256": OpCode(
            "OP_SHA256",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": BTCVector.__hash__("sha256"),
                "symbol": "sha256",
            },
        ),
        "OP_CODESEPARATOR": OpCode("OP_CODESEPARATOR", instruction_constructor=NOP),
        "OP_CHECKSIG": OpCode(
            "OP_CHECKSIG",
            instruction_constructor=CheckSigInstruction,
            default_args={
                "sig_function": BTCVector.check_sig,
                "constraint_on_operands": signature_constraints,
            },
        ),
        "OP_CHECKMULTISIG": OpCode(
            "OP_CHECKMULTISIG",
            instruction_constructor=CheckMultiSigInstruction,
            default_args={"sig_function": BTCVector.check_multisig},
        ),
        # expansion
        "OP_NOP1": OpCode("OP_NOP1", instruction_constructor=NOP),
        "OP_CHECKLOCKTIME": OpCode(
            "OP_CHECKLOCKTIME",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": BTCVector.__identity__,
                "symbol": "check_lock_time",
                "constraint_on_operands": locktime_constraint,
            },
        ),
        "OP_CHECKSEQUENCE": OpCode(
            "OP_CHECKSEQUENCE",
            instruction_constructor=ArithmeticOperator,
            default_args={
                "arity": 1,
                "operator": BTCVector.__identity__,
                "symbol": "check_sequence",
                "constraint_on_operands": sequence_constraint,
            },
        ),
        "OP_NOP4": OpCode("OP_NOP4", instruction_constructor=NOP),
        "OP_NOP5": OpCode("OP_NOP5", instruction_constructor=NOP),
        "OP_NOP6": OpCode("OP_NOP6", instruction_constructor=NOP),
        "OP_NOP7": OpCode("OP_NOP7", instruction_constructor=NOP),
        "OP_NOP8": OpCode("OP_NOP8", instruction_constructor=NOP),
        "OP_NOP9": OpCode("OP_NOP9", instruction_constructor=NOP),
        "OP_NOP10": OpCode("OP_NOP10", instruction_constructor=NOP),
        # TODO implement Opcode added by BIP 342 (Tapscript)
        "OP_CHECKSIGADD": OpCode(
            "OP_CHECKSIGADD",
            instruction_constructor=InvalidInstruction,
            default_args={
                "token": "OP_CHECKSIGADD",
                "tag": InstructionTag.NOT_IMPLEMENTED,
            },
        ),
        "OP_INVALIDOPCODE": OpCode(
            "OP_INVALIDOPCODE",
            instruction_constructor=InvalidInstruction,
            default_args={"token": "OP_INVALIDOPCODE", "tag": InstructionTag.RESERVED},
        ),
        # /!\ Not part of BTC protocol. Custom defined operator.
        # Thus, we do not need to add a integer representation for this opcode.
        # No script will ever contain a JUMP opcode.
        "OP_JUMP": OpCode(
            "OP_JUMP", branching=True, instruction_constructor=UnconditionalJump
        ),
    }

    def int_to_tokens(opcode: int) -> str:
        """Return the token(s) associated to an opcode (int).

        To reduce the number of operators that need to be implemented, an
        opcode can return several tokens.

        Examples:
            * OpCodes._opcode_to_token(0x88)
                >> "OP_EQUAL OP_VERIFY"
            * OpCodes._opcode_to_token(0x9d)
                >> "OP_NUMEQUAL OP_VERIFY"
        """
        if opcode in OpCodes.__int_to_tokens:
            return OpCodes.__int_to_tokens[opcode]
        else:
            raise InvalidOpCode(f"No known token for {hex(opcode)}")

    def token_to_opcode(token: str) -> OpCode:
        """Return the Opcode (object) associated to a token."""
        if token in OpCodes.__token_to_op_codes:
            return OpCodes.__token_to_op_codes[token]
        else:
            return None

    def valid_tokens(filter_branching=True) -> List[str]:
        """Return the valid tokens.

        Set `filter_branching` to `True` to list only the non branching tokens.
        """
        tokens = []
        for token, op_code in OpCodes.__token_to_op_codes.items():
            if not filter_branching or not op_code.branching:
                tokens.append(token)
        return tokens

    def valid_integers() -> List[int]:
        """Return the set of integers associated to tokens."""
        return list(OpCodes.__int_to_tokens.keys())
