"""This class implements a parser for the Bitcoin script language."""
from math import inf
from typing import Dict, List, Tuple
from pyparsing import (
    Literal,
    hexnums,
    ZeroOrMore,
    countedArray,
    Optional,
    Or,
    Word,
    Combine,
    Forward,
    OneOrMore,
    Optional,
    MatchFirst,
    Keyword,
    Group,
)
import pyparsing
from symbolic_execution.language.exceptions import IncompleteIfEndIfClause

pyparsing.ParserElement.enablePackrat()
from collections import deque
from symbolic_execution.program.block.interfaces import IInstructionBlock

from symbolic_execution.program.block.instruction_block import InstructionBlock
from symbolic_execution.program.instruction.stack_based import PushDataOperator
from ..interfaces import ILanguage
from .op_codes import OpCodes
import sys

sys.setrecursionlimit(1000000)

def read_little_endian(hex: str, bytes: int, start: int = 0) -> int:
    """Read and returns the integer value of a hexadecimal value in little endian contained in a string.

    Args:
        hex: A string containing a hexadecimal value encoded in little endian.
        bytes: The number of bytes to read.
        start: The offset of the beginning of the hex value.

    Returns:
        int: The integer value of the hex string in little endian.

    Examples:
        read_little_endian('xxxaf12xx', 2, 3) will return 4783
    """
    ordered = [hex[i : i + 2] for i in range(start, start + bytes * 2, 2)][::-1]
    string = "".join(ordered)
    return int(string, 16)

class BitcoinParser(ILanguage):
    """Parser for the Bitcoin script language."""

    def tokenize_code(self, code: str) -> List[str]:
        """Parse {code} & returns the tokens."""
        return BitcoinParser.__tokenify_script(code)

    def build_control_flow_graph(self, tokens: List[str]) -> IInstructionBlock:
        """Parse the tokens and return the control flow graph."""
        if tokens.count(OpCodes.int_to_tokens(99)) != tokens.count(
            OpCodes.int_to_tokens(104)
        ):
            raise IncompleteIfEndIfClause("")
        control_flow_graph = BitcoinParser.__build_control_flow_graph(tokens)
        BitcoinParser.__set_block_pc(control_flow_graph)
        BitcoinParser.__build_instructions(control_flow_graph)
        return control_flow_graph

    def __tokenify_script(script: str) -> str:
        """First pass on the script to produce tokens and parse the data.

        Example: __tokenify_script("63020007674c040102030468") will return:
        >> "OP_IF 0007 OP_ELSE 01020304 OP_ENDIF"
        """
        # General concepts
        BYTE = Word(hexnums, exact=2)
        # The only difference with a BYTE, is that an OP_CODE needs to be
        # converted into a token.
        OP_CODE = Word(hexnums, exact=2)

        @OP_CODE.addParseAction
        def tokenify(tokens):
            return list(map(lambda t: OpCodes.int_to_tokens(int(t, 16)), tokens))

        # CONSTANTS
        OP_PUSH_0 = MatchFirst([format(op, "02x") for op in range(0x01, 0x4C)])
        OP_PUSH_0.addParseAction(lambda t: int(t[0], 16))
        OP_PUSH_1 = Literal("4c") + Word(hexnums, exact=2)
        OP_PUSH_1.addParseAction(lambda t: read_little_endian(t[1], 1))
        OP_PUSH_2 = Literal("4d") + Word(hexnums, exact=4)
        OP_PUSH_2.addParseAction(lambda t: read_little_endian(t[1], 2))
        OP_PUSH_4 = Literal("4e") + Word(hexnums, exact=8)
        OP_PUSH_4.addParseAction(lambda t: read_little_endian(t[1], 4))
        OP_PUSH = MatchFirst([OP_PUSH_0, OP_PUSH_1, OP_PUSH_2, OP_PUSH_4])
        DATA = Combine(countedArray(BYTE, intExpr=OP_PUSH))

        SCRIPT = Combine(ZeroOrMore(MatchFirst([DATA, OP_CODE])), joinString=" ")

        return SCRIPT.parseString(script)[0]

    def __build_control_flow_graph(tokenified_script: str):
        """Build the control flow graph of the script."""
        # Special instruction blocks that stand for unsuccessful execution.
        # Every opcode that can raise an exception (e.g. OP_VERIFY) points to
        # this block.
        FINAL_BLOCK = InstructionBlock([], final=True)
        # This block is optional and is only used with OP_RETURN and OP_VERIFY
        FAILED_EXECUTION_BLOCK = InstructionBlock(["00"])

        def __init_failed_block(failed_block):
            """Init the failed block as a parent to the final block."""
            if not failed_block.children:
                failed_block.add_children([FINAL_BLOCK])
            return failed_block

        # PARSER
        NON_FLOW_OPCODES = MatchFirst(
            [
                Word(hexnums),
                *list(map(lambda token: Keyword(token), OpCodes.valid_tokens())),
            ]
        )

        CONSECUTIVE_INSTRUCTIONS = OneOrMore(NON_FLOW_OPCODES)
        BLOCK = Forward()
        BLOCK << (
            MatchFirst(
                [
                    (
                        Optional(CONSECUTIVE_INSTRUCTIONS)("BLOCK_A")
                        + (Keyword("OP_IF"))("BRANCHING_KW")
                        + Optional(BLOCK)("BLOCK_B")
                        + ZeroOrMore(
                            Group(Keyword("OP_ELSE") + Optional(BLOCK)("BLOCK_C"))
                        )("ELSE_CLAUSES")
                        + Keyword("OP_ENDIF")
                        + Optional(BLOCK)("BLOCK_D")
                    ),
                    (
                        Optional(CONSECUTIVE_INSTRUCTIONS)("BLOCK_A")
                        + Keyword("OP_VERIFY")("OP_VERIFY")
                        + Optional(BLOCK)("BLOCK_B")
                    ),
                    (
                        Optional(CONSECUTIVE_INSTRUCTIONS)("BLOCK_A")
                        + Keyword("OP_RETURN")("OP_RETURN")
                        + Optional(BLOCK)("UNREACHABLE_BLOCK")
                    ),
                    CONSECUTIVE_INSTRUCTIONS,
                ]
            )
        )
        PROGRAM = Optional(BLOCK)

        # Parse actions
        @CONSECUTIVE_INSTRUCTIONS.addParseAction
        def group_instructions(tokens):
            return InstructionBlock(tokens)

        @BLOCK.addParseAction
        def build_graph(tokens):
            if "BRANCHING_KW" in tokens:
                blockA = (
                    tokens["BLOCK_A"][0] if "BLOCK_A" in tokens else InstructionBlock()
                )
                if_clause = (
                    tokens["BLOCK_B"] if "BLOCK_B" in tokens else InstructionBlock()
                )
                else_clauses = (
                    list(
                        map(
                            lambda else_clause: else_clause["BLOCK_C"]
                            if "BLOCK_C" in else_clause
                            else InstructionBlock(),
                            tokens["ELSE_CLAUSES"],
                        )
                    )
                    if len(tokens["ELSE_CLAUSES"]) > 0
                    else [InstructionBlock()]
                )

                blockD = (
                    tokens["BLOCK_D"] if "BLOCK_D" in tokens else InstructionBlock()
                )

                blockA.add_elts([tokens["BRANCHING_KW"]])

                blockA.add_children([if_clause, else_clauses[0]])

                # Chain consecutive else clauses together
                blocks_to_chain = (
                    [if_clause]
                    + [else_clause for else_clause in else_clauses[1::2]]
                    + [blockD]
                )
                for block, next_block in zip(blocks_to_chain[:-1], blocks_to_chain[1:]):
                    block.add_children([next_block])
                blocks_to_chain = [
                    else_clause for else_clause in else_clauses[0::2]
                ] + [blockD]
                for block, next_block in zip(blocks_to_chain[:-1], blocks_to_chain[1:]):
                    block.add_children([next_block])

            elif "OP_VERIFY" in tokens:
                blockA = (
                    tokens["BLOCK_A"][0] if "BLOCK_A" in tokens else InstructionBlock()
                )
                blockB = (
                    tokens["BLOCK_B"] if "BLOCK_B" in tokens else InstructionBlock()
                )

                blockA.add_elts([tokens["OP_VERIFY"]])
                blockA.add_children(
                    [blockB, __init_failed_block(FAILED_EXECUTION_BLOCK)]
                )
            elif "OP_RETURN" in tokens:
                blockA = (
                    tokens["BLOCK_A"][0] if "BLOCK_A" in tokens else InstructionBlock()
                )
                # Let's drop the unreachable block.
                # We don't need it for symbolic execution.
                if "UNREACHABLE_BLOCK" in tokens:
                    unreachable_block = tokens["UNREACHABLE_BLOCK"]
                    unreachable_block.disconnect_block()

                blockA.add_elts([tokens["OP_RETURN"]])
                blockA.add_children([__init_failed_block(FAILED_EXECUTION_BLOCK)])
            else:
                # This case corresponds to a simple block of instructions
                return tokens

            return [blockA]

        @PROGRAM.addParseAction
        def add_final(tokens):
            if len(tokens) == 0:
                return [FINAL_BLOCK]
            elif len(tokens) == 1:
                tokens[0].add_children([FINAL_BLOCK])
            else:
                raise ValueError(
                    "Should not happen normally. A program is composed of one or zero "
                    "block."
                )

            return tokens

        parsed_code = PROGRAM.parseString(tokenified_script)

        return parsed_code[0]

    def __set_block_pc(control_flow_graph: InstructionBlock):
        """
        First pass on the control flow graph.

        * Set the program counters for every block of instructions.
        * Set a JUMP op code at the end of every block if there's not
        already a (conditionnal or unconditonnal) JUMP.
        """
        fifo_queue = deque(
            [control_flow_graph]
        )  # used for a breadth first serach over the graph
        visited = set()

        block_counter = 0  # pc of the first instruction in the current block
        # First pass to set the pc of the first instruction in every block
        while len(fifo_queue) > 0:
            block = fifo_queue.pop()
            # Do not revisit the same block
            if block in visited:
                continue

            visited.add(block)
            block.pc = block_counter

            for c in block.children:
                fifo_queue.appendleft(c)

            # Every block should finish with a jump (cond. or not).
            # Add another token 'JUMP' if the final block's
            # token is not a branching token
            last_opcode = (
                OpCodes.token_to_opcode(block.elts[-1]) if len(block.elts) > 0 else None
            )

            if not last_opcode or not last_opcode.branching:
                block.add_elts(["OP_JUMP"])

            block_counter += len(block.elts)

    def __build_instructions(control_flow_graph: InstructionBlock):
        """Build the instructions of the script.

        Parse the control flow graph of the script and replaces the tokens
        with real instructions.
        """
        fifo_queue = deque(
            [control_flow_graph]
        )  # used for a breadth first serach over the graph
        visited = set()
        instructions = {}

        while len(fifo_queue) > 0:
            block = fifo_queue.pop()
            if block in visited:
                continue

            visited.add(block)
            block_instructions = []

            for i, token in enumerate(block.elts):
                pc = block.pc + i

                if token.startswith("OP_"):
                    generator = OpCodes.token_to_opcode(token).instruction_constructor
                    if not generator:
                        raise ValueError(
                            f"""No generator implemented for this token:
                            {token}"""
                        )

                    default_args = OpCodes.token_to_opcode(token).default_args
                    if len(block.children) > 0:
                        destination_pcs = list(map(lambda c: c.pc, block.children))
                    else:
                        destination_pcs = [inf]

                    instr = generator(
                        pc=pc, **{**default_args, "destination_pcs": destination_pcs}
                    )

                else:
                    instr = PushDataOperator(pc=pc, data=token)

                block_instructions.append(instr)
                instructions[instr.pc] = instr

            # Replace the tokens stored inside the block by the instructions.
            block.reset_elts()
            block.add_elts(block_instructions)

            for c in block.children:
                fifo_queue.appendleft(c)

        return instructions
