"""The code presented here aims at trying to find a valid unlocking script."""

from enum import Enum
from multiprocessing import Manager, Process, Queue
from symbolic_execution.hardware.interfaces import IMachine
from symbolic_execution.hardware.machines.btc_machine import (
    CHECK_SINGLE_SIG,
    BTCMachine,
    BTC_VECTOR_SIZE,
    HASH_FUNCTIONS,
)
from symbolic_execution.hardware.memory.ram import FullySymbolicRAM
from symbolic_execution.language.bitcoin.parser import BitcoinParser
from symbolic_execution.language.exceptions import (
    IncompleteIfEndIfClause,
    InvalidOpCode,
)
from symbolic_execution.program.executer.symbolic import SymbolicExecuter
from symbolic_execution.program.instruction.exceptions import (
    UnimplementedInstructionError,
)
from symbolic_execution.program.instruction.general import InstructionTag
from z3 import *
import numpy as np
from symbolic_execution.program.block.utils import compute_path_count

class Satisfiability(Enum):
    SATISFIABLE = 0
    UNSATISFIABLE = 1
    UNKNOWN = 2
    UNSPENDABLE = 3

    @classmethod
    def z3_to_enum(cls, satisfiability):
        if satisfiability == sat:
            return cls.SATISFIABLE
        elif satisfiability == unsat:
            return cls.UNSATISFIABLE
        elif satisfiability == unknown:
            return cls.UNKNOWN
        else:
            raise ValueError("Unknown z3 satifiablility constant.")

    def __str__(self) -> str:
        return self.name.lower()

    def __lt__(self, other):
        return str(self) < str(other)


def __find_unlocking_script_timed(hex_script, return_dict, verbose, count_path):
    m = ""
    satisfiability = unknown

    try:
        btc_parser = BitcoinParser()
        tokens = btc_parser.tokenize_code(hex_script)

        if verbose:
            print(tokens)
        program = btc_parser.build_control_flow_graph(tokens)
        if verbose:
            print(program)

        machine = BTCMachine(FullySymbolicRAM, BTC_VECTOR_SIZE, stack_count=2)

        symbolic_executer = SymbolicExecuter()
        final_machine = symbolic_executer.execute(program, machine)
        top_elt = final_machine.pop_from_stack()
        if verbose:
            print(top_elt.val)

        s = Solver()
        s.add(top_elt.val != 0)
        s.add(final_machine._constraints)
        z3.set_param("model.compact", False)

        satisfiability = s.check()
        if satisfiability == sat:
            m = s.model()

        return_dict["satisfiability"] = Satisfiability.z3_to_enum(satisfiability)
        return_dict["model"] = str(m)
        if count_path:
            return_dict["path_count"] = compute_path_count(program)

    except InvalidOpCode as e:
        return_dict["satisfiability"] = Satisfiability.UNSPENDABLE
        return_dict["model"] = {}

    except IncompleteIfEndIfClause as e:
        return_dict["satisfiability"] = Satisfiability.UNSPENDABLE
        return_dict["model"] = {}

    except UnimplementedInstructionError as e:
        return_dict["satisfiability"] = Satisfiability.UNSPENDABLE
        return_dict["model"] = {}

    except Exception as e:
        raise e

    finally:
        if count_path and "path_count" not in return_dict:
            return_dict["path_count"] = np.nan


def find_unlocking_script(hex_script: str, timeout=None, verbose=False, count_path=False):
    """Parse, execute the script. Then, it tries to find a valid unlocking script."""
    manager = Manager()
    return_dict = manager.dict()

    p1 = Process(
        target=__find_unlocking_script_timed, args=(hex_script, return_dict, verbose, count_path)
    )
    p1.start()
    p1.join(timeout=timeout)
    p1.terminate()

    if p1.exitcode is None:
        return "", Satisfiability.z3_to_enum(unknown), np.nan
    elif count_path:
        return return_dict["model"], return_dict["satisfiability"], return_dict["path_count"]
    else:
        return return_dict["model"], return_dict["satisfiability"], np.nan