"""Instruction block and multiple executer implementations."""
from typing import Generic, TypeVar
from ..block.interfaces import IInstructionBlock

Block = TypeVar("Block")


class InstructionBlock(IInstructionBlock, Generic[Block]):
    """Represent a block of consecutive instructions.

    Two instructions {i_1} and {i_2} are said to be consecutive, if the
    execution of {i_1} necessarily induces the execution of {i_2}.

    In other words, {i_1} can't alter the execution flow. Note that there is no
    restriction on {i_2}

    A program is made of several instruction blocks linked together.
    """

    def add_children(self, children: list):
        """Chain another blocks to the current one.

        If the current block has already children, {children} are added to them.
        """
        if self._final:
            return

        # Remove self from the list of children to prevent loops
        children = list(filter(lambda child: child is not self, children))

        if self._children:
            for c in self._children:
                c.add_children(children)
        else:
            super().add_children(children)

    def least_common_ancestor(self, other: Block):
        """Return the least common ancestor between {self} and {other}.

        Return None if no ancestor can be found.
        """
        self_ancestors = self.__ancestors()
        other_ancestors = other.__ancestors()
        common_ancestors = {
            key: self_ancestors[key]
            for key in self_ancestors.keys() & other_ancestors.keys()
        }

        return min(common_ancestors, key=common_ancestors.get)

    def __ancestors(self, n=0):
        if self._parents:
            ancestors = {parent: n + 1 for parent in self._parents}
            for parent in self._parents:
                ancestors.update(parent.__ancestors(n + 1))
        else:
            ancestors = {self: n}
        return ancestors
