"""Interfaces for instruction blocks."""
from abc import ABC, abstractmethod
from typing import List, TypeVar, Generic


Block = TypeVar("Block")
T = TypeVar("T")


class IInstructionBlock(ABC, Generic[Block]):
    """Represent a block containing a list of elements.

    Every block may be chained with some other blocks.

    Example: A program is made of several instruction blocks linked together.
    """

    @property
    def pc(self):
        """Return the program counter of the first instruction of this block."""
        return self._pc

    @pc.setter
    def pc(self, pc):
        self._pc = pc

    @property
    def children(self):
        """Return the blocks following the current block."""
        return self._children

    @property
    def parents(self):
        """Return the blocks preceeding the current block."""
        return self._parents

    @property
    def elts(self):
        """Return the instructions contained in the block."""
        return self._elts

    def __init__(self, elts: List[T] = None, final: bool = False):
        """Initialize a block with the values it contains and its children.

        The block can be set to final to prevent any child to be added.
        """
        self._elts = elts if elts else []
        self._children = []
        self._parents = []
        self._final = final

        self._pc = 0

    def add_elts(self, elts: List[T]):
        """Extend the list of elements of the block with {elts}."""
        self._elts.extend(elts)

    def reset_elts(self):
        """Reset all the elements stored within the block."""
        self._elts.clear()

    @abstractmethod
    def add_children(self, children: List[Block]):
        """Chain current block with the other blocks {children}.

        The implementation is left to the user regarding the use case, e.g. to ensure no
        cycle exist.
        """
        if self._final:
            return

        self._children.extend(children)
        for c in children:
            c._add_parent(self)

    def disconnect_block(self):
        """Disconnect the current block from all children and parents."""
        for child in self._children:
            child._parents.remove(self)
        for parent in self._parents:
            parent._children.remove(self)

    def _add_parent(self, parent: Block):
        """Chain current block with another other block {Block}."""
        self._parents.extend([parent])

    @abstractmethod
    def least_common_ancestor(self, other: Block) -> Block:
        """Return the least common ancestor between {self} and {other}.

        Return None if no ancestor can be found.

        The implementation is left to the user, because the custom implementation of
        `add_children` might have an impact on the algorithm employed here.
        """
        pass

    def __str__(self, tab=""):
        """Return the string representation of the block and its children."""
        s = self.__repr__() + "\n"
        for c in self._children:
            s += f"{tab}|\n{tab}-> " + c.__str__(tab=tab + "  ")
        return s

    def __repr__(self):
        """Representation of the block."""
        s = f"BLOCK["
        for elt in self._elts:
            s += f"{elt} "
        s += f"]"
        return s
