"""Implementation of a concrete executer of a program."""

from collections import deque
from typing import Dict

from symbolic_execution.program.block.interfaces import IInstructionBlock

from ..block.instruction_block import InstructionBlock
from symbolic_execution.hardware.interfaces import IMachine, ISymbolicMemory
from ..executer.interfaces import IExecuter
from ..instruction.interfaces import IInstruction


class ConcreteExecuter(IExecuter):
    """Concrete executer of a program."""

    def build_flat_program(self, block: IInstructionBlock) -> Dict[int, IInstruction]:
        """Convert the graph of blocks to a flat program."""
        # Build the program as a mapping between the program counters and the
        # instructions.
        program = {}
        visited = set()
        fifo_queue = deque([block])

        while len(fifo_queue) > 0:
            block = fifo_queue.pop()
            if block not in visited:
                visited.add(block)

                # Check there's no pair of instruction sharing the same pc
                block_pcs = set([instruction.pc for instruction in block.elts])
                program_pcs = set(program.keys())
                pcs_collisions = program_pcs.intersection(block_pcs)
                if len(pcs_collisions) > 0:
                    raise ValueError(
                        f"Several pcs are shared by different instructions: "
                        "{pcs_collisions}"
                    )

                program.update(
                    {instruction.pc: instruction for instruction in block.elts}
                )

            for c in block.children:
                fifo_queue.appendleft(c)

        return program

    def execute(self, block: InstructionBlock, machine: IMachine):
        """Execute the program in a concrete fashion.

        For example, the execution will abort if the condition of a branching operator
        can't be evaluated to `True` or `False`.
        """
        program = self.build_flat_program(block)

        machine.pc = 0
        try:
            while machine.pc < len(program) and machine.pc >= 0:
                instr = program[machine.pc]
                instr.run(machine)

            # Pop and push again to keep the stack as it is.
            if (
                isinstance(machine._stacks[0], ISymbolicMemory)
                or machine.count_stack_elements() > 0
            ):
                top_elt = machine.pop_from_stack()
                machine.push_onto_stack(top_elt)

                return top_elt, None
            else:
                return False, None

        except Exception as e:
            return False, e
