"""Instructions for any type of machine."""
from enum import Enum
from typing import Any, List

from .exceptions import UnimplementedInstructionError, InstructionInitializationError
from .interfaces import IInstruction, IUnconditionalBranchingOperator

from symbolic_execution.hardware.interfaces import IMachine, IRAMBasedMachine


class InstructionTag(Enum):
    """Reason why an instruction is not implemented."""

    NOT_IMPLEMENTED = 0
    RESERVED = 1
    DISABLED = 2

    def __str__(self):
        """Return the name in lowercase."""
        return f"{self.name}".lower()


class NOP(IInstruction):
    """Instruction that does nothing."""

    def __init__(self, pc: int, **args):
        """Initialize a the instruction."""
        super().__init__(pc)

    def run(self, machine: IMachine):
        """Run the operator on any machine."""
        super().run(machine)

    def __repr__(self) -> str:
        """Return the string representation of the instruction."""
        return f"{super().__repr__()}()"


class UnconditionalJump(IUnconditionalBranchingOperator):
    """Instruction that modifies the pc."""

    def __init__(self, pc: int, destination_pcs: List[int], **args):
        """Initialize a the instruction."""
        super().__init__(pc, **args)

        if len(destination_pcs) != 1:
            raise InstructionInitializationError(
                f"{self.__class__.__name__}: Expects exactly 1 possible "
                "destination pc, while the argument passed "
                "is {destination_pcs}."
            )

        self._destination_pcs = destination_pcs

    def run(self, machine: IMachine):
        """Run the operator on any machine."""
        super().run(machine)

        machine.pc = self._destination_pcs[0]

    def __repr__(self) -> str:
        """Return the string representation of the instruction."""
        return f"{super().__repr__()}({self._destination_pcs[0]})"


class InvalidInstruction(IInstruction):
    """Instruction that raises an exception when run.

    Raise a `UnimplementedInstructionError` exception.
    """

    def __init__(
        self,
        pc: int,
        token: str,
        tag: InstructionTag = InstructionTag.NOT_IMPLEMENTED,
        **args,
    ):
        """Init an instruction that raises an exception when run."""
        super().__init__(pc, **args)
        self._token = token
        self._tag = tag

    def run(self, machine: IMachine):
        """Raise an UnimplementedInstructionError exception."""
        msg = f"""The opcode {self._token} at program position {self._pc} stands \
for a {self._tag} instruction."""
        raise UnimplementedInstructionError(msg, self._tag)

    def __repr__(self) -> str:
        """Return the string representation of the instruction."""
        return f"{super().__repr__()}({self._token})"
