"""Instrution for hybrid types of machine.

E.g. the BTC machine works mostly as a stack, but also allow random memory access.

Note: some of this instructions are hardly portable across different languages.
"""
from typing import Any, Callable, Type
from symbolic_execution.hardware.interfaces import (
    IMachine,
    IRAMBasedMachine,
    IStackBasedMachine,
    ISymbolicMachine,
)
from symbolic_execution.program.instruction.exceptions import (
    InstructionInitializationError,
)
from .interfaces import IInstruction
from z3 import *


class PickInstruction(IInstruction):
    """Read the data x hops from the top and push onto the stack."""

    def __init__(self, pc: int, **args):
        """Init the instruction."""
        super().__init__(pc, **args)

    def run(self, machine: Type[IMachine]):
        """Return the value in memory at position {pos} along with its size."""
        if not isinstance(machine, IStackBasedMachine) and not isinstance(
            machine, IRAMBasedMachine
        ):
            raise TypeError(
                "This instruction needs to run on a machine that is both a stack based "
                "and a ram based machine."
            )

        super().run(machine)

        n = machine.pop_from_stack()
        elt_n = machine.read(n)
        machine.push_onto_stack(elt_n)

    def __repr__(self) -> str:
        """Return the string representation of the instruction."""
        return f"{super().__repr__()}(top_stack_elt)"


class CheckMultiSigInstruction(IInstruction):
    """IInstruction that checks several signatures against public keys."""

    def __init__(
        self,
        pc: int,
        sig_function: Callable,
        constraint_on_operands: Callable = None,
        **args,
    ):
        """Initialize the check sig instruction."""
        super().__init__(pc, **args)

        if sig_function is None:
            raise InstructionInitializationError("The signature function is None.")
        self._sig_function = sig_function

        self._constraint_on_operands = constraint_on_operands

    def run(self, machine: IStackBasedMachine):
        """Check several signatures against several public keys."""
        super().run(machine)

        pubkeys_count = machine.pop_from_stack()
        if is_int_value(pubkeys_count.val):
            pubkeys = [
                machine.pop_from_stack() for _ in range(pubkeys_count.val.as_long())
            ]
        else:
            machine.drop_from_stack(pubkeys_count)
            pubkeys = []

        sig_count = machine.pop_from_stack()
        if is_int_value(sig_count.val):
            sigs = [machine.pop_from_stack() for _ in range(sig_count.val.as_long())]
        else:
            machine.drop_from_stack(sig_count)
            sigs = []
        # Due to the bug in the BTC implementation, an extra element is dropped
        machine.pop_from_stack()

        ret = self._sig_function(pubkeys_count, sig_count, pubkeys, sigs)
        machine.push_onto_stack(ret)

        if self._constraint_on_operands and isinstance(machine, ISymbolicMachine):
            constraints = self._constraint_on_operands([pubkeys, sigs])
            machine.add_constraints(constraints)

    def __repr__(self) -> str:
        """Return the string representation of the instruction."""
        return f"{super().__repr__()}({'multi_sig'})"
