import itertools
import unittest
from symbolic_execution.program.instruction.interfaces import IInstruction

from symbolic_execution.language.bitcoin.op_codes import OpCodes
from .constants import VALID_INTEGERS, VALID_TOKENS


class TestOpcodes(unittest.TestCase):
    btc_language = OpCodes

    def test_int_to_tokens_coverage(self):
        """Checks there's a token for every possible integer in a script."""
        for i in VALID_INTEGERS:
            with self.subTest(i=i):
                tokens = self.btc_language.int_to_tokens(i)
                self.assertIsNotNone(
                    tokens, f"The integer `{hex(i)}` has no corresponding token."
                )

    def test_token_to_opcode_coverage(self):
        """Checks there's an opcode for every token defined."""
        all_valid_integers = self.btc_language.valid_integers()

        for i in all_valid_integers:
            # As a recall, an integer can be associated to multiple tokens
            tokens = self.btc_language.int_to_tokens(i)

            for token in tokens.split(" "):
                with self.subTest(f"{i}: {token}"):
                    op_code = self.btc_language.token_to_opcode(token)
                    self.assertIsNotNone(
                        op_code,
                        f"The integer `{i}` is associated to the token `{token}`, "
                        "which in turn, has no corresponding opcode.",
                    )

    def test_token_has_an_int(self):
        """Checks there's a valid integer for every token defined."""
        tokens_with_opcode = set(self.btc_language.valid_tokens(filter_branching=False))
        # This token is not expected to have a corresponding integer, because we
        # defined it and it's not part of BTC language.
        tokens_with_opcode.remove("OP_JUMP")

        all_valid_integers = self.btc_language.valid_integers()
        tokens_with_int = set(
            itertools.chain(
                *list(
                    map(
                        lambda i: self.btc_language.int_to_tokens(i).split(" "),
                        all_valid_integers,
                    )
                )
            )
        )

        for token_with_op_code in tokens_with_opcode:
            with self.subTest(f"{token_with_op_code}"):
                self.assertIn(
                    token_with_op_code,
                    tokens_with_int,
                    f"This token {token_with_op_code} is "
                    "defined, but there's no integer that can produce it.",
                )

    def test_valid_integers(self):
        """Test the valid integers function."""
        self.assertEqual(
            self.btc_language.valid_integers(),
            VALID_INTEGERS,
            "Regression test failed. Something has changed in the function "
            "`valid_integers`",
        )

    def test_valid_non_branching_tokens(self):
        """Regression test against the `valid_tokens` function, only non_branching."""
        valid_non_branching_tokens = list(
            map(
                lambda token: token.name,
                filter(lambda token: not token.is_branching, VALID_TOKENS),
            )
        )

        self.assertEqual(
            self.btc_language.valid_tokens(),
            valid_non_branching_tokens,
            "Regression test failed.  Something has changed in the function"
            "`valid_tokens`",
        )

    def test_valid_tokens(self):
        """Regression test against the `valid_tokens` function."""
        self.assertEqual(
            self.btc_language.valid_tokens(filter_branching=False),
            list(map(lambda token: token.name, VALID_TOKENS)),
            "Regression test failed.  Something has changed in the function"
            "`valid_tokens`",
        )

    def test_token_instruction_constructor(self):
        """Test that every valid token has a valid generator for an instruction."""
        valid_tokens = self.btc_language.valid_tokens(filter_branching=False)
        for token in valid_tokens:
            with self.subTest(f"{token}"):
                op_code = self.btc_language.token_to_opcode(token)

                self.assertIsNotNone(
                    op_code, f"The opcode for this token {token} is " "set to None."
                )

                gen = op_code.instruction_constructor
                default_args = op_code.default_args

                self.assertIsNotNone(
                    gen, f"The generator for this token {token} is not initialized."
                )

                if gen is not None:
                    # Generate the instruction with mock arguments
                    destination_pcs = (
                        [0] if token in ["OP_JUMP", "OP_RETURN"] else [0, 1]
                    )
                    instruction = gen(
                        pc=0, **{**default_args, "destination_pcs": destination_pcs}
                    )
                    self.assertIsInstance(instruction, IInstruction)


if __name__ == "__main__":
    unittest.main()
