from collections import deque
import copy
import unittest
from symbolic_execution.hardware.interfaces import ISymbolicMemory
from symbolic_execution.hardware.memory.ram import FullySymbolicRAM

from symbolic_execution.language.bitcoin.parser import BitcoinParser
from symbolic_execution.hardware.machines.btc_machine import BTC_VECTOR_SIZE, BTCMachine
from symbolic_execution.program.executer.concrete import ConcreteExecuter
from .concrete.testing_scenarii import TESTING_SCENARII as CONCRETE_TESTING_SCENARII
from .symbolic.testing_scenarii import TESTING_SCENARII as SYMBOLIC_TESTING_SCENARII
from symbolic_execution.language.bitcoin.symbolic_interpreter import (
    find_unlocking_script,
)
from z3 import is_expr


class TestParser(unittest.TestCase):
    btc_parser = BitcoinParser()

    def test_tokenify_script(self):
        for scenario in CONCRETE_TESTING_SCENARII:
            with self.subTest(scenario.name):
                tokens = self.btc_parser.tokenize_code(scenario.hex_script)
                self.assertEqual(tokens, scenario.tokens)

    def test_build_program(self):
        for scenario in CONCRETE_TESTING_SCENARII:
            with self.subTest(scenario.name):
                graph = self.btc_parser.build_control_flow_graph(scenario.tokens)
                program = ConcreteExecuter().build_flat_program(graph)
                self.assertEqual(str(list(program.values())), scenario.program)

    def test_concrete_execution(self):
        for memory_type in [FullySymbolicRAM]:
            for scenario in CONCRETE_TESTING_SCENARII:
                with self.subTest(f"{scenario.name} on {memory_type.__name__}"):
                    graph = self.btc_parser.build_control_flow_graph(scenario.tokens)

                    machine = BTCMachine(memory_type, BTC_VECTOR_SIZE, stack_count=2)
                    status, error = ConcreteExecuter().execute(graph, machine)
                    if error:
                        raise error

                    # In the case of concrete execution, a depleted stack at the end of
                    # the execution will be considered as unsuccessful.
                    # At the contrary, for symbolic execution, no assumption can be
                    # made, i.e. it would be possible to add an extra element in the
                    # input script would make the execution valid or invalid.
                    if scenario.execution_status is not None:
                        self.assertEqual(
                            status,
                            scenario.execution_status,
                            "The return status of the execution is different from what "
                            f"was expected. Err raised during the execution: \n{error}",
                        )

                    expected_final_stack_states = copy.deepcopy(
                        scenario.final_stack_states
                    )

                    for stack_n, final_state in enumerate(expected_final_stack_states):
                        # Does not check the size of the stack if the memory is
                        # symbolic
                        if not issubclass(memory_type, ISymbolicMemory):
                            self.assertEqual(
                                machine.count_stack_elements(n=stack_n),
                                len(final_state),
                                "The number of elements on the stack is different "
                                "from the expected number of elements.",
                            )

                        elt_i = 0
                        while len(final_state):
                            expected_top_elt = final_state.pop()
                            top_elt = machine.pop_from_stack(n=stack_n)
                            top_elt = (
                                str(top_elt.val)
                                if isinstance(expected_top_elt, str)
                                else top_elt.val
                            )

                            self.assertEqual(
                                top_elt,
                                expected_top_elt,
                                f"The element #{elt_i} from the top of the stack "
                                "differs from what was expected.",
                            )

                            elt_i += 1

    def test_symbolic_execution(self):
        for _, scenario in enumerate(SYMBOLIC_TESTING_SCENARII):
            with self.subTest(scenario.hex_script):
                _, status, path_count = find_unlocking_script(scenario.hex_script, timeout=30)
                self.assertEqual(status, scenario.execution_status)


if __name__ == "__main__":
    unittest.main()
